#define _CRT_SECURE_NO_WARNINGS
#include <stdio.h>
#include <Windows.h>
#include "hooks.h"

#define ENOMEM 12

static void *hook_malloc(unsigned int size);
static void *hook_calloc(unsigned int nmemb, size_t size);
static void *hook_realloc(void *ptr, size_t size);

struct func_hook hooks[] = {
		{ "msvcr100.dll", "malloc", (unsigned long)hook_malloc, 0 },
		{ "msvcr100.dll", "calloc", (unsigned long)hook_calloc, 0 },
		{ "msvcr100.dll", "realloc", (unsigned long)hook_realloc, 0 }
	};


static int calls_limit[sizeof(hooks) / sizeof(struct func_hook)];

static char *GetImportByOrdName(HANDLE hModule, DWORD impOrd)
{
	IMAGE_DOS_HEADER *dosHdr;
	IMAGE_NT_HEADERS *peHdr;
	ULONG loadBase;
	IMAGE_DATA_DIRECTORY *dir;
	IMAGE_EXPORT_DIRECTORY *expDir;
	DWORD i;
	SHORT *AddressOfNameOrdinals;
	DWORD *AddressOfNames;

	dosHdr = (IMAGE_DOS_HEADER *)hModule;
	loadBase = (ULONG)dosHdr;

	peHdr = (IMAGE_NT_HEADERS *)(loadBase + dosHdr->e_lfanew);
	dir = &peHdr->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT];

	expDir = (IMAGE_EXPORT_DIRECTORY *)(loadBase + dir->VirtualAddress);

	AddressOfNameOrdinals = (SHORT *)(loadBase + expDir->AddressOfNameOrdinals);
	AddressOfNames = (DWORD *)(loadBase + expDir->AddressOfNames);

	for (i = 0; i < expDir->NumberOfNames; i++) {

		if (AddressOfNameOrdinals[i] == (impOrd - expDir->Base))
			return (char *)(loadBase + AddressOfNames[i]);
	}

	return NULL;
}

static void PatchAddr(DWORD_PTR *addr, DWORD_PTR val)
{
	BOOL bRet;
	DWORD oldProtect;
	DWORD dummy;
	MEMORY_BASIC_INFORMATION mbi;

	VirtualQuery(addr, &mbi, sizeof(mbi));

	bRet = VirtualProtect(mbi.BaseAddress, mbi.RegionSize, PAGE_READWRITE, &oldProtect);
	if (!bRet) {
		printf("VirtualProtect failed: %d\n", GetLastError());
		return;
	}

	*addr = val;

	bRet = VirtualProtect(mbi.BaseAddress, mbi.RegionSize, oldProtect, &dummy);
	if (!bRet) {
		printf("VirtualProtect failed: %d\n", GetLastError());
		return;
	}
}

int install_hooks(char *target_lib_name, struct func_hook *hooks, int num)
{
	IMAGE_DOS_HEADER *dosHdr;
	IMAGE_NT_HEADERS *peHdr;
	IMAGE_DATA_DIRECTORY *dir;
	DWORD *oft;
	DWORD *ft;
	IMAGE_IMPORT_BY_NAME *impName;
	DWORD impOrd;
	HANDLE hModule;
	char *dllName;
	char *func_name;
	IMAGE_IMPORT_DESCRIPTOR *impDesc;
	ULONG_PTR loadBase;
	int i;

	dosHdr = (IMAGE_DOS_HEADER *)GetModuleHandle(target_lib_name);
	if (!dosHdr)
		return FALSE;

	loadBase = (ULONG)dosHdr;

	peHdr = (IMAGE_NT_HEADERS *)(loadBase + dosHdr->e_lfanew);

	dir = &peHdr->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT];

	for (i = 0; i < num; i++) {
		impDesc = (IMAGE_IMPORT_DESCRIPTOR *)(loadBase + dir->VirtualAddress);

		while(impDesc->OriginalFirstThunk) {
			dllName = (char *)(loadBase + impDesc->Name);

			if (!_stricmp(dllName, hooks[i].lib_name)) {
				hModule = GetModuleHandle(dllName);
				if (!hModule) {
					fprintf(stderr, "Couldn't get handle to %s: %d\n", dllName, GetLastError());
					return FALSE;
				}

				oft = (DWORD *)(loadBase + impDesc->OriginalFirstThunk);
				ft = (DWORD *)(loadBase + impDesc->FirstThunk);
				for (; *oft; oft++, ft++) {
					if (*oft & 0x80000000) {
						impOrd = *oft;
						impOrd &= 0xffff;

						func_name = GetImportByOrdName(hModule, impOrd);

						if (func_name && !_stricmp(func_name, hooks[i].func_name)) {
							hooks[i].orig_addr = (unsigned long)GetProcAddress(hModule, hooks[i].func_name);
							PatchAddr(ft, hooks[i].addr);
						}
					} else {
						impName = (IMAGE_IMPORT_BY_NAME *)(loadBase + *oft);
						func_name = (char *)impName->Name;

						if (!_stricmp(func_name, hooks[i].func_name)) {
							hooks[i].orig_addr = (unsigned long)GetProcAddress(hModule, hooks[i].func_name);
							PatchAddr(ft, hooks[i].addr);
						}
					}
				}
			}

			impDesc++;
		}
	}
	return 0;
}


static void *hook_malloc(size_t size)
{
	static int nr_times;
	void *(*orig_malloc)(size_t);
	int malloc_limit = calls_limit[0];

	orig_malloc = (void *(*)(size_t))hooks[0].orig_addr;

	if (malloc_limit == -1) {
		return orig_malloc(size);
	}

	if (nr_times >= malloc_limit) {
		errno = ENOMEM;
		return NULL;
	}

	nr_times++;

return orig_malloc(size);
}

static void *hook_calloc(size_t nmemb, size_t size)
{
	static int nr_times;
	void *(*orig_calloc)(size_t, size_t);
	int calloc_limit = calls_limit[1];

	orig_calloc = (void *(*)(size_t, size_t))hooks[1].orig_addr;

	if (calloc_limit == -1) {
		return orig_calloc(nmemb, size);
	}

	if (nr_times >= calloc_limit) {
		errno = ENOMEM;
		return NULL;
	}

	nr_times++;

	return orig_calloc(nmemb, size);
}

static void *hook_realloc(void *ptr, size_t size)
{
	static int nr_times;
	void *(*orig_realloc)(void *, size_t);
	int realloc_limit = calls_limit[2];

	orig_realloc = (void *(*)(void *, size_t))hooks[2].orig_addr;

	if (realloc_limit == -1) {
		return orig_realloc(ptr, size);
	}

	if (nr_times >= realloc_limit) {
		errno = ENOMEM;
		return NULL;
	}

	nr_times++;

	return orig_realloc(ptr, size);
}

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD	 fdwReason, LPVOID	lpvReserved)
{
	if (fdwReason == DLL_PROCESS_ATTACH) {
		int i;

		/* The order from env_vars and hooks must be the same */
		char *env_vars[] = {
			"malloc_limit",
			"calloc_limit",
			"realloc_limit"
		};

		int nr_hooks = sizeof(env_vars) / sizeof(env_vars[0]);

		for (i = 0; i < nr_hooks; i++) {
		char *env_var_val = getenv(env_vars[i]);
		if (env_var_val != NULL) {
			sscanf(env_var_val, "%d", &calls_limit[i]);
		} else {
			calls_limit[i] = -1;
		}
		}
		install_hooks(NULL, hooks, nr_hooks);
	}

	return TRUE;
}
